def version = "1.3.0"
setInfos(9, "Chastity Combination", "Keyholding with a combination lock and timer. (Demonstrator)", "tinyinsect", "v"+version, 0x001bb1, "en", ["chastity", "forfemale", "formale", "keyholding"]);

// This script, from https://ss.deviatenow.com, is protected by this licence :
// CC by-NC, see https://creativecommons.org/licenses/by-sa/4.0/

// 0.1 First public testing with restrictions
// 1.0 Unrestricted working version
//	- non-unique, longer, and alpha-numeric combinations
//	- hygiene lockup for 15 minutes every 12 hours
//	- Choose random duration
//	- Hide the lock time
//	- Lock info in the start screen when running
//	- Show the combination with monospace font
// 1.0.1 Fixes
//	- Reduced the number for distraction digits and failed tests
// 1.1.0 Update
//	- Show combination characters within hidden combination (better to identify the character position)
// 1.2.0 Update
//	- Add storage of the combination on the server: Online settings, ID handling, send and receive functions
//	- Add combination display with fake characters in random order
// 1.3.0 Update
//	- Use custom images in image directory if they are present

def currentversion = loadString("ChastityCombination.version")
// show version update notes once
setImage(null)
switch (currentversion) {
	case null: //notes for versions >1.2.0, no number saved before
		show("Update notes:\n\n"+
			"To change the image of your keyholder, put one or more jpeg-files into the image directory of the script (<tt>images/ChastityCombination</tt>). They will be randomly selected and shown.")
		showButton("OK")
// 	case "1.3.0": // updates for >1.3.0
}
save("ChastityCombination.version", version)

def testing = false
def debug = false
def debugmsg = { text ->
	if (debug) {
		show(text)
		showButton("OK")
	}
	return
}

def encodeB64 = { String input ->
	// base64 encoding of input string
	// returns string
	return input.bytes.encodeBase64().toString()
}

def decodeB64 = { String input ->
	// base64 decoding of input string
	// returns string
	if (input) {
		byte[] decoded = input.decodeBase64()
		return new String(decoded)
	}
	else return input
}

def shufflestr = { String input ->
	// random shuffling of input string
	// returns string and string list of indices of original character position
	def strlen = input.length()
	def strshuffled = ""
	def indxstr = ""
	//list of all indices
	def indices = (0..strlen-1).collect()
	for (def i=0; i<strlen; i++) {
		def indx = indices.remove(getRandom(indices.size()))
		strshuffled += input[indx]
		indxstr+=(String.format("%02d", indx))
		}
	return [strshuffled, indxstr]
}

def reorderstr = { String input, indices ->
	// order the characters in the input string according to the indices
	// indices: string of indices (2 digits)
	def strlen = input.length()
	def strordrd = []*strlen
	for (def i=0; i<strlen; i++) {
		def indx = indices[i*2..(i*2)+1].toInteger()
		strordrd[indx] = input[i]
		}
	return strordrd.join()
}

def encode = { String input ->
	// encoding of input string
	// returns string
	def shuffled = shufflestr(input)
	return encodeB64(shuffled.join())
}

def decode = { String input ->
	// decoding of input string
	// returns string
	if (input) {
		def plainstr = decodeB64(input)
		def pl = plainstr.length()
		return reorderstr(plainstr[0..pl/3-1], plainstr[-(2*pl/3)..pl-1])
	}
	else return input
}

def randstrgenerator = { String alphabet, int n, unique=false ->
	// returns a string of length n with random characters from alphabet
	// with unique characters if set to true. Then n must not be larger than len(alphabet)!
	if (unique) {
		if (n > alphabet.length()) showPopup("Error: cannot create string of that length with unique characters. Ignoring n.")
		def randstr = ""
		def imax = Math.min(n, alphabet.length())
		for (def i=0; i<imax; i++) {
			def randchar = alphabet[getRandom(alphabet.length())]
			alphabet -= randchar
			randstr += randchar
		}
		return randstr
	}
	else new Random().with {
			(1..n).collect { alphabet[ nextInt( alphabet.length() ) ] }.join()
			}
}

def randomstr = { n, chartype=3, unique=false ->
	// returns a string with n characters of the following type:
	// chartype 1: only digits (0..9)
	// 	2: digits and lowercase characters (0..9, a--z)
	//	3 (default): digits, lowercase, and uppercase characters (0..9, a--z, A..Z)
	// with unique characters if set to true
	// returns empty string for invalid chartype
	def charstr = ""
	if (chartype >= 1) charstr += ('0'..'9').join()
	if (chartype >= 2) charstr += ('a'..'z').join()
	if (chartype >= 3) charstr += (('A'..'Z').join() - "O") //remove O in favor of 0
	return randstrgenerator(charstr, n, unique)
}

def padstr = { String str, length=16, chartype=3 ->
	// pad string with random characters on the left to length total characters
	// string input must have less than 100 characters, otherwise the inout is returned unchanged
	// chartype 1: only digits (0..9)
	// 	2: digits and lowercase characters (0..9, a--z)
	//	3 (default): digits, lowercase, and uppercase characters (0..9, a--z, A..Z)
	// will ignore length if it is too small and add no padding
	// returns string
	//   [padding]+[str]+[len(str)](2 digits)
	def strlen = str.length()
	if (strlen > 99) {showPopup("Error: cannot pad string longer than 99 characters!"); return str}
	def npad = length - strlen - 2
	def padstr = ""
	if (npad > 0) padstr += randomstr(npad, chartype)
	return padstr + str + String.format("%02d", strlen)
}

def stripstr = { str ->
	// extract the string which was padded with padnumber()
	// input string [padding]+[string]+[len(string)](2 digits)
	// returns string
	if (str) {
		def strlen = str[-2..-1] as int
		return str[-2-strlen..-3]
	}
	else return str
}

def padnumber = { number, length=16 ->
	// pad int or long number with random digits on the left to length total digits
	// will ignore length if it is too small and add no padding
	// returns string
	//   [padding]+[number]+[len(number)](2 digits)
	return padstr(number.toString(), length, 1)
}

def tttxt = { str, big=false ->
	// return string with monospace formatting
	// big font when set to true
	if (big) return "<tt><big>${str}</big></tt>"
	else return "<tt>${str}</tt>"
// 	return "<span style=\"font-family: monospace;\">${str}</span>"
}

def showComboSingle = { String combo, cdisplay=0, chartype=1 ->
	// show combination string in single characters
	// cdisplay: way to display the characters
	//	0: only single characters one by one
	//	1: combination in *, single characters unveiled in order
	//	2: combination in *, single characters unveiled in random order
	//	3: combination in *, single characters unveiled in random order, 1--3 fake iterations before
	// chartype: character type for fake iteration, default: numbers
	if (cdisplay==0) {
		for (def i=0; i<combo.length(); i++) {
			show(null)
			wait(0.2)
			show("${tttxt(combo[i], true)}")
			showButton("OK")
		}
	}
	else if (cdisplay in [1,2,3]) {
		def combostr = ""
		def cl = combo.length()
		def extrachar = 0
		if (cdisplay == 3) extrachar = (Math.ceil((23-cl)/3)-1 as int) - 2 + getRandom(3)
		def indices = (0..cl-1).collect()
		if (cdisplay in [2,3]) Collections.shuffle(indices)
		def index = 0
		def cchar = ""
		for (def i=0; i<(cl+extrachar); i++) {
			combostr = "*" * combo.length()
			show("${tttxt(combostr, true)}")
			wait(0.3)
			if (i >= extrachar) {
				index = indices[i-extrachar]
				cchar = combo[index]
			}
			else {
				index = getRandom(cl)
				cchar = randomstr(1, chartype)
			}
			combostr = "*"*index + cchar + "*"*(cl-index-1)
			show("${tttxt(combostr, true)}")
			showButton("OK")
		}		
	}
	else showPopup("Error: Invalid value for cdisplay: $cdisplay")
	show(null)
}

def sectotxt = { seconds ->
	// convert number in seconds to string format ((d days,) h hours and )m minutes
	def d = Math.floor(seconds/3600/24) as int
	def h = Math.floor((seconds-d*3600*24)/3600) as int
	def m = Math.floor((seconds-d*3600*24-h*3600)/60) as int
	return "${(d>0?"${d} days, ":"")}${(h>0?"${h} hours and ":"")}${m} minutes"
}

def getID = { svrcheck=false ->
	// create ID with string characters
	// if svrcheck: check for collision online
	// then, if no connection is present return null
	//		if no free ID is found 10x return -1
	if (svrcheck && !isConnected()) return null
	def id
	def i=0
	while (true) {
		id = randomstr(7, 2)
		for (def j=0; j<2; j++)
			id += "-" + randomstr(7, 2)
		if (svrcheck) {
			def idonline = receiveString(id)
			if (!idonline) break
		}
		else break
		if (i==10) return -1
		i++
	}
	return id
}

// some personalization to make it a nice script. Defaults to Mistress/boy/cock.
def isfemale = loadBoolean("intro.female")
def likemale = loadBoolean("intro.likemale")
def genital = (isfemale?"pussy":"cock")
def slave = (isfemale?"girl":"boy")
def keyholder = (likemale?"Master":"Mistress")
// images: check for custom ones, or use default one
def imgfilelist = new File("${getDataFolder()}images/ChastityCombination").listFiles()
def imagefiles = imgfilelist.findAll { f -> f.name.endsWith(".jpg"); }.name
imagefiles += imgfilelist.findAll { f -> f.name.endsWith(".jpeg"); }.name
if (imagefiles.size() == 0) {
	if (likemale) {
		imagefiles += ["../default-male.jpg"]
	}
	else {
		imagefiles += ["../default-female.jpg"]
	}
}
def setNewImage = { first=false ->
	if (first || imagefiles.size() > 1) {
		setImage("ChastityCombination/"+imagefiles[getRandom(imagefiles.size())])
	}
}

// load saved lock and parameters
def csettings = stripstr(decode(loadString("ChastityCombination.CA")))
def ccombo = stripstr(decode(loadString("ChastityCombination.CC")))
def cbegin = stripstr(decode(loadString("ChastityCombination.CB")))
def cduration = stripstr(decode(loadString("ChastityCombination.CD")))
def c911 = stripstr(decode(loadString("ChastityCombination.CE")))
def cllast = ""
def cid = loadString("ChastityCombination.CI")
// check if a lock is there and still running
boolean lockexists
int cremain
boolean lockrunning
def getLockState = {
	// set variables for the lock status: exists, remaining, running
	lockexists = cbegin && cduration
	if (lockexists) {
		cremain = ((cbegin as int) + (cduration as int)) - getTime() // in seconds
	}
	else cremain = 0
	lockrunning = cremain > 0
	cllast = stripstr(decode(loadString("ChastityCombination.CH")))
}
def unlock = {
	// erase cleaning, lock start and duration, but keep the combination
	cbegin = null
	cduration = null
	cllast = null
	save("ChastityCombination.CB", cbegin)
	save("ChastityCombination.CD", cduration)
	save("ChastityCombination.CH", cllast)
}

// combination parameters:
// length, type, recurring
// combination length, 2..12
int clength = 4
// combination type
//	1: only digits (0..9)
// 	2: digits and lowercase characters (0..9, a--z)
//	3: digits, lowercase, and uppercase characters (0..9, a--z, A..Z)
int ctype = 1
// unique characters in the combination
boolean cunique = false
// hide the remaining time
def chidden = false
// store combination on the server: true for first run without ID
def conline = true
if (csettings) {
	// load saved parameters if possible
	if (csettings.length() < 5) csettings = "0" + csettings //default chidden: false
	if (csettings.length() < 6) csettings = "1" + csettings //default conline: true
	conline = (csettings[0]=="1"?true:false)
	chidden = (csettings[1]=="1"?true:false)
	ctype = csettings[2] as int
	cunique = (csettings[3] =="1"?true:false)
	clength = csettings[4..5] as int
}
def ctypes = ["numerical digits", "digits and lowercase letters", "digits, lower- and uppercase letters"]
def saveSettings = {
	//save the configuration
	csettings = 1e5*(conline?1:0) + 1e4*(chidden?1:0) + 1e3*ctype + 1e2*(cunique?1:0) + clength
	save("ChastityCombination.CA", encode(padnumber(csettings)))
}

// server synchronization status
def sync = loadBoolean("ChastityCombination.sync")
def setSync = {status ->
	sync = status
	save("ChastityCombination.sync", sync)
}
// trailing chars for the server to avoid cut of leading zeros of numbers. Must not be in original string!
def idsuffix = "-??-"
def sendData = {
	if (isConnected()) {
		send(cid, ccombo + idsuffix) //add trailing chars
		return true
	}
	else return false
}
def receiveCombo = {
	return receiveString(cid) - idsuffix //remove trailing chars
}

def lockstart = false // lockup at the start of a lock
def jumpto = null // alternative target block if needed

// Here we go
setNewImage(true)
show("Hello slave!")
wait(2)

// testing only. promise! no: harsh exit
if (testing) {
	show("<big>WARNING!\n\n"+
			"This script is not finalized and may malfunction or be altered!\n\n"+
			"Promise me to NOT use it for an actual lockup!</big>")
	wait(10)
	if (!getBoolean(null, "I promise!", "Why not?")) {
		show("It is too risky at this stage. I can't guarantee that the script works as intended and you may not be able to unlock.")
		wait(5)
		if (!getBoolean(null, "OK, I promise!", "I don't care")) {
			show("Sorry, then I cannot do anything for you!")
			wait(5)
			setImage(null)
			show(null)
			return null
		}
	show("Good ${slave}! I don't want you to get yourself in danger!")
	showButton("Thanks, ${keyholder}")		
	}
}

def text = ""
def block = "start"
//main loop
def endReached = false
def newblock = block
while(!endReached) {
	block = null
	switch(newblock) {
		case "start":
			// check if an emergency key has been commanded, also checks for the first start
			if (!c911) {
				setNewImage()
				if (!getBoolean("I can be your keyholder for your ${genital}. All you need is a chastity device, and a combination safe or lock box to store the keys.\n\n"+
						"I will give you a combination for it and hide it from you until time is over.", "Yes, please, ${keyholder}", "No, thanks")) {
					show("As you want. Goodbye!")
					wait(3)
					block = ""
					break
				}	
				show("Before we start, remember that computers may break, code may be buggy, data may get corrupted. Do not rely exclusively on me to unlock you!\n\n"+
						"<big>Therefore I order you to store an emergency key accessible somewhere without needing this combination!</big>\n\n"+
						"Secure it with a plastic lock or seal it in an envelope, but do it now!")
				wait(10)
				if (getBoolean(null, "Done as told, ${keyholder}", "No I won't")) {
					show("Good ${slave}! I don't want you to get yourself in danger! Play safe!")
					showButton("Yes, ${keyholder}")	
				}
				else {
					show("If you don't listen, I will not do anything for you!")
					wait(5)
					block = ""
					break
				}
				save("ChastityCombination.CE", encode(padnumber(1)))
			}
			// check if online storage is enabled but no ID is present: ask for online storage of combination
			if (!cid && conline) {
				if (!getBoolean("Do you want me to store the combination on the server as a backup?", "Yes, please, ${keyholder}", "No, thanks")) {
					show("No problem, but watch your data. You can enable it later.")
					conline = false
					// save settings: online data
					saveSettings()
					showButton("OK")
				}
				else {
					// start subroutine to generate ID and store the combination if there is one
					conline = true
					// save settings: online data
					saveSettings()
					jumpto = "main"
					block = "newID"
					break
				}
			}
			block = "main"
			break
		
		case "main":
			// main start pages
			setNewImage()
			debugmsg("$conline, $chidden, $ctype, ${ctypes[ctype-1]}, $cunique, $clength, ${csettings}, $sync")
			// initialize lock state
			getLockState()
			debugmsg("exist ${lockexists}, running ${lockrunning}, remain ${cremain}\n"+
					"combo ${ccombo}, cbegin ${cbegin}, cduration ${cduration}, cllast ${cllast}\n"+
					"cid $cid")
			// try to sync with server if not completed
			if (conline && !sync) {
				if (debug) showPopup("Trying to send data")
				setSync(sendData())
			}
			if (lockrunning) {
				block = "running"
			}
			else {
				if (lockexists) {
					block = "over"
				}
				else {
					block = "unlocked"
				}
			}
			break
			
		case "unlocked": 
			// new lock, show last combination, or exit
			text = "You are currently unlocked. However, you belong in chastity and you know it!\n\n"+
					"What do you want to do?"
			def selections = ["Start new lock"]
			if (ccombo) selections.add("Show last combination")
			if (conline) selections.add("Check online data")
			else  selections.add("Set online data")
			selections.add("Exit")
			def blocks = ["newlock"]
			if (ccombo) blocks.add("showcombo")
			blocks.add("onlinedata")
			blocks.add("exit")
			block = blocks[getSelectedValue(text, selections)]
			break
		
		case "onlinedata":
			//en/disable online data storage, get online combination, enter ID
			setNewImage()
			text = "Online storage is ${conline?"enabled":"disabled"}.\n"+
					"${conline?"Online synchronization was ${!sync?"NOT":""} done.\n":""}"+
					"${(conline && cid)?"Your ID is: ${tttxt(cid)}\n":""}\n"+
					"Select an option:"
			def selections = []
			if (!conline) selections.add("Enable online storage")
			selections.add("Show online combination")
			if (cid && !conline) selections.add("Show ID")
			if (conline) selections.add("Set new ID")
			if (conline) selections.add("Disable online storage")
			selections.add("Back")
			def blocks = []
			if (!conline) blocks.add("startonline")
			blocks.add("onlinecombo")
			if (cid && !conline) blocks.add("showonlineID")
			if (conline) blocks.add("newonlineID")
			if (conline) blocks.add("stoponline")
			blocks.add("back")
			def section = blocks[getSelectedValue(text, selections)]
			switch (section) {
				case "startonline":
					conline = true
					// save settings: online data
					saveSettings()
					if (!cid) {
						if (getBoolean("There is no ID yet, do you want to generate a new one, or enter an existing one?", "New ID", "Existing ID")) block = "newID"
						else block = "enterID"
					}
					else block = "onlinedata"
					break
				case "onlinecombo":
					block = "loadcombo"
					break
				case "showonlineID":
					block = "showID"
					break
				case "newonlineID":
					if (getBoolean("Do you want to generate a new one, or enter an existing one?", "New ID", "Existing ID")) block = "newID"
					else block = "enterID"
					break
				case "stoponline":
					conline = false
					// save settings: online data
					saveSettings()
					block = "onlinedata"
					break
				case "back":
					block = "main"
					break
			}
			break

		case "newID":
			// generate ID, store combination if existent
			if (cid) {
				if (!getBoolean("Warning: your current ID will be replaced! Do you want to continue?")) {
					block = "onlinedata"
					break
				}
			}
			def useonline = true
			cid = getID(useonline)
			while (!cid) {
				if (getBoolean("Error: Cannot connect to the server, make sure you are online. Try again?")) cid = getID(useonline)
				else break
				show(null)
				wait(0.2)
			}
			while (cid == -1) {
				if (getBoolean("Cannot generate a free ID... Try again?")) cid = getID(useonline)
				else break
			}
			if (cid && cid != -1) {
				show("This is your ID:\n\n${tttxt(cid)}\n\n"+
						"Note it down, keep it secret and somewhere safe! You will need it to retrieve the combination from the server if you lose your data.")
				wait(10)
				showButton("ID noted")
				save("ChastityCombination.CI", cid)
				show("You can also always look up your ID in the online data if you need to.")
				showButton("OK")
			}
			else {
				show("Sorry, I did not succeed to create a usable ID. You may try again later, but for now I cannot store the combination online.")
				showButton("Understood, $keyholder")
				conline = false
				// save settings: online data
				saveSettings()
			}
			block = jumpto
			if (!block) block = "onlinedata"
			jumpto = null
			break

		case "enterID":
			//manually enter the online ID and save it if wanted
			def oldid = cid
			def sampleid = getID()
			def abort = false
			while (!abort) {
				cid = getString((oldid?"The current ID is:\n\n${tttxt(oldid)}\n\n":"")+
					"Enter your known ID:", "")
				if (cid.length() != sampleid.length()) {
					switch (getSelectedValue("The ID ${tttxt(cid)} has not the expected length. Try again, use it anyway, or abort?", ["Again", "Use this one", "Abort"])) {
						case 1:
							abort = true
							break
						case 2:
							cid = oldid
							abort = true
							break
					}
					
				}
				else break
			}
			if (cid != oldid) {
				if (oldid) {
					if (getBoolean("Do you want to save ${tttxt(cid)} as the new ID?")) {
						save("ChastityCombination.CI", cid)
						show("ID saved")
						wait(2)
					}
				}
				else {
					save("ChastityCombination.CI", cid)
					show("ID saved")
					wait(2)
				}
			}
			else if (!cid) {
				show("Online storage is not possible without an ID. Disabling it.")
				conline = false
				// save settings: online data
				saveSettings()
				showButton("OK")
			}
			block = jumpto
			if (!block) block = "onlinedata"
			jumpto = null
			break
		
		case "loadcombo":
			//load combination from server: with existing or user ID
			if (cid) {
				def isonline = isConnected()
				while (!isonline) {
					if (!getBoolean("Error: Cannot connect to the server, make sure you are online. Try again or abort?", "Again", "Abort")) {
						block = "onlinedata"
						break
					}
					isonline = isConnected()
					show(null)
					wait(0.2)
				}
				if (isonline) {
					def servercombo = receiveCombo()
					text = "The used ID is: ${tttxt(cid)}\n\n"
					if (servercombo) show(text + "The combination on the server is:\n\n${tttxt(servercombo)}\n\n"+
								"Do you want to manually enter an ID?")
					else show(text + "I cannot find a combination on the server for it... "+
								"Do you want to manually enter an ID?")
				}
			}
			else show("I cannot find an ID, do you want to manually enter it?")
			if (!block) {
				if (getBoolean(null)) {
					jumpto = "loadcombo"
					block = "enterID"
				}
				else {
					block = "onlinedata"
				}
			}
			break
		
		case "showID":
			show("Your ID is:\n\n${tttxt(cid)}")
			showButton("Thank you, ${keyholder}")
			block = jumpto
			if (!block) block = "onlinedata"
			jumpto = null
			break
			
		case "over": 
			// lock over: unlock or exit
			def cpassed = getTime() - (cbegin as int)
			text = "Congratulations! Your time in chastity is over!\n\n"+
					"You have been locked for ${sectotxt(cpassed)}.\n\n"+
					"What do you want to do?"
			def selections = ["Unlock"]
			if (cid) selections.add("Show ID")
			selections.add("Exit")
			def blocks = ["unlock"]
			if (cid) blocks.add("showID")
			blocks.add("exit")
			block = blocks[getSelectedValue(text, selections)]
			if (block == "showID") jumpto = "over"
			break

		case "running": 
			// show time. cleaning if allowed, or emergency unlock, or exit
			// check if the required time has passed for cleaning, default OK as no cllast may be present
			def clallow = true
			int clwait
			if (cllast) {
				clwait = 12*3600 - (getTime() - (cllast as int))
				clallow = clwait <= 0
			}
			text = "You are currently in chastity, as you should be!\n\n"
			text += "You have been locked for ${sectotxt(getTime()-(cbegin as int))}.\n"
			if (chidden) text += "The remaining time remains my secret.\n\n"
			else text += "The remaining time of your lock is ${sectotxt(cremain)}.\n\n"
			if (!clallow) text += "You have to wait ${sectotxt(clwait)} until your next cleaning unlock.\n\n"
			text += "What do you want to do?"
			def selections = []
			if (clallow) selections.add("Cleaning unlock")
			if (cid) selections.add("Show ID")
			selections.add("Emergency unlock")
			selections.add("Exit")
			def blocks = []
			if (clallow) blocks.add("cleanunlock")
			if (cid) blocks.add("showID")
			blocks.add("911unlock")
			blocks.add("exit")
			block = blocks[getSelectedValue(text, selections)]
			if (block == "showID") jumpto = "running"
			break

		case "unlock":
			// show combination
			setNewImage()
			show("Alright slave, the combination of your lock is:\n\n${tttxt(ccombo)}\n\n"+
					"You may unlock yourself now.")
			wait(10)
			showButton("Unlocked")
			unlock()
			show("I hope you've had as much fun as I did.")
			showButton("Yes, ${keyholder}, thank you!")
			block = "main"
			break

		case "showcombo":
			// show combination
			setNewImage()
			show("Sure! The combination of your last lock is:\n\n${tttxt(ccombo)}")
			wait(2)
			showButton("Thank you, ${keyholder}")
			block = "unlocked"
			break

		case "911unlock":
			// emergency unlock
			setNewImage()
			// ask for confirmation
			if (!getBoolean("You need to free your ${genital}?\n\nBe honest: is it a real emergency?")) {
				setNewImage()
				show("Then why do you ask? You belong in chastity more than you think!")
				wait(7)
				block = ""
				break
			}
			setNewImage()
			show("OK, I trust you on this one. The combination of your lock is:\n\n${tttxt(ccombo)}\n\n"+
					"You may unlock yourself now.")
			wait(10)
			showButton("Unlocked")
			unlock()
			show("I hope you are fine. Remember to always keep an emergency key ready, I may not be able to let you out all the time.")
			showButton("Yes, ${keyholder}, thank you!")
			block = "main"
			break

		case "newlock":
			// start new lock
			setNewImage()
			show("Good decision, let's get started!")
			wait(3)
			setNewImage()
			show("First and foremost:\nGet your chastity device and lock away this ${genital}!")
			wait(15)
			showButton("Locked, ${keyholder}")
			setNewImage()
			show("Have you stored the emergency key as requested?")
			wait(2)
			if (!getBoolean(null, "I have, ${keyholder}", "No, and I won't")) {
				show("If you don't listen, I will not do anything for you!")
				wait(5)
				setImage(null)
				show(null)
				return null
			}
			// hide remaining time
			chidden = getBoolean("Should I hide the lock duration from you?")
			// ask duration
			if (getBoolean("Do you want to have a fixed or random lockup time?", "Fixed", "Random")) {				
				while (true) {
					cduration = getInteger("How many hours do you want to be locked up?", 0) * 3600 //in seconds
					if (testing && cduration > 24*3600) {
						show("Sorry, right now I don't allow more than a day. I will however once I know everything works.")
						showButton("OK")
					}
					else if (cduration > 720*3600) {
						if (getBoolean("Are you sure you want to be locked for ${Math.round(cduration/24/3600)} days?")) break
					}
					else if (cduration > 0) break
				}
			}
			else {
				int cdmin
				int cdmax
				while (true) {
					while (true) {
						cdmin = getInteger("Enter the minimum time for the lock in hours:", 0) * 3600 //in seconds
						if (cdmin >= 0) break
					}
					while (true) {
						cdmax = getInteger("And the maximum time for the lock in hours:", (cdmin/3600 as int)+1) * 3600 //in seconds
						if (cdmax < cdmin) {
							show("The maximum time should be larger than the minimum, silly ${slave}!")
							showButton("Sorry")
						}
						else break
					}
					if (testing && cdmax > 24*3600) {
						show("Sorry, right now I don't allow the maximum to be more than a day. I will however once I know everything works.")
						showButton("OK")
					}
					else if (cdmax > 30*24*3600) {
						if (getBoolean("Are you sure you want to be locked for up to ${Math.round(cdmax/24/3600)} days?")) break
					}
					else break
				}
				cduration = cdmin + getRandom(cdmax+1 - cdmin)
				if (chidden) show("OK, I picked your time in chastity. You will see when it is over...")
				else show("OK, I decided to keep you in chastity for ${sectotxt(cduration)}.")
				showButton("Thank you, ${keyholder}")
			}
			// save duration
			save("ChastityCombination.CD", encode(padnumber(cduration)))
			// show parameters, ask for confirmation
			setNewImage()
			while (getBoolean("I will give you a combination of the following type:\n\n"+
						"${clength} ${(cunique?"unique":"non-unique")} ${ctypes[ctype-1]}\n\n"+
						"Do you want to change any of these parameters?")) {
				// change combination parameters
				while (true) {
					while (true) {
						clength = getInteger("What combination length do you want? Choose between 2 and 14.", clength)
						if (clength > 14 || clength < 2) {
							show("I said between 2 and 14!")
							showButton("OK, sorry")
						}
						else break
					}
					ctype = getSelectedValue("What kind of combination do you want?\n\n"+
							"1: only numbers (${tttxt("0--9")})\n"+
							"2: numbers and lowercase letters (${tttxt("0--9, a--z")})\n"+
							"3: numbers, upper- and lowercase letters (${tttxt("0--9, A--Z, a--z")})", ["1: (0--9)", "2: (0--9, a--z)", "3: (0--9, A--Z, a--z)"]) + 1
					cunique = getBoolean("Should each ${(ctype>1?"character":"number")} be unique in the combination?")
					// check for too long combination
					if (cunique && ((ctype == 1 && clength > 10) || (ctype == 2 && clength > 36) || (ctype == 3 && clength > 81))) {
						show("A combination of this length and type is not possible with unique characters! Choose other parameters!")
						cunique = false
						showButton("OK")
					}
					else break
				}
			}
			// save settings: new lock parameters
			saveSettings()
			// create the lock
			lockstart = true
			block = "lockup"
			break
			
		case "lockup":	
			// create combination, lock the key
			// check if a connection is possible if needed
			def isonline = false
			if (conline) {
				isonline = isConnected()
				while (!isonline) {
					if (!getBoolean("Error: Cannot connect to the server, storing the combination online is not possible. Try again now or work offline and try later?", "Again", "Offline")) break
					isonline = isConnected()
					show(null)
					wait(0.2)
				}
			}
			setNewImage()
			show("I will now show you the combination with hidden ${(ctype>1?"characters":"digits")} and reveal one by one. You set each one on the lock, and click the button each time for the next one.\n\n"+
			"Should I show them in normal order, random order, or random with fake ${(ctype>1?"characters":"digits")}?")
			def cdisplay = getSelectedValue(null, ["Normal", "Random", "Random with fakes"]) + 1
			while (true) {
				ccombo = randomstr(clength, ctype, cunique)
				// show combination
				show("Get ready")
				showButton("Ready")
				showComboSingle(ccombo, cdisplay, ctype)
				if (!getBoolean("Do you want to have a different combination?")) break
			}
			show(null)
			wait(0.3)
			showButton("Lock set")
			// save combo local and online (plain text)
			save("ChastityCombination.CC", encode(padstr(ccombo, 16, ctype)))
			if (conline) setSync(sendData())
			else setSync(false)
			// show distraction
			setNewImage()
			show("OK. I will you show you some more ${(ctype>1?"characters":"digits")} to distract you. Memorize them or write them down, I will question you!")
			showButton("OK, ready")
			def cfake
			def i=0
			while (i<3) {
				while (true) {
					cfake = randomstr(7-ctype, ctype, cunique)
					if (cfake != ccombo) break
				}
				showComboSingle(cfake, 1)
				if (getString("So tell me the ${(ctype>1?"characters":"digits")} I showed.", "")!=cfake) {
					text = "Wrong"
					if (i<2) {
						text += "! Concentrate and try again!"
						show(text)
						showButton("OK")
					}
					else {
						text += "... but you tried."
						show(text)
						wait(2)
						break
					}
				}
				else {
					text = "Correct"
					if (i<2) {
						text += "! And once more..."
						show(text)
						i=1
						wait(3)
					}
					else {
						text += "... that's enough."
						show(text)
						wait(2)
						break
					}
				}
				i++
				setNewImage()
			}
			// finalize the lock
			setNewImage()
			show("Now store the key and close the combination lock.")
			wait(10)
			showButton("Key is locked")
			if (lockstart) {
				setNewImage()
				cbegin = getTime()
				save("ChastityCombination.CB", encode(padnumber(cbegin)))
				show("Good ${slave}, that's it! Now your ${genital} belongs to me, and I will only tell you the combination again when time is up.\n\n"+
						"You may however briefly unlock for cleaning every 12 hours, and you can ask for an emergency release if you REALLY need to.")
				showButton("Thank you, ${keyholder}!")
				lockstart = false
			}
			else {
				setNewImage()
				show("OK, we are done here. Stay good and chaste for me!")
				showButton("Thank you, ${keyholder}!")
			}				
			block = "main"
			break

		case "cleanunlock":
			// hygiene unlock
			setNewImage()
			def maxcltime = 15 //allowed time in minutes
			show("Hygiene is important! You may unlock for cleaning for $maxcltime minutes.\n\n"+
					"I will show you the combination and expect you back before time is up. Be aware that you get 1 hour added on your chastity duration per 5 minute delay!")
			showButton("Understood, ${keyholder}")
			setNewImage()
			show("The combination of your lock is:\n\n${tttxt(ccombo)}\n\n"+
					"Go clean yourself and your device now and report back when done. Don't touch your ${genital} more than needed!")
			wait(60)
			def maxwait = (maxcltime-2)*60
			def cltime = showButton("Clean and back", maxwait)
			if (cltime == maxwait) {
				playBackgroundSound("bell.wav")
				setNewImage()
				show("Hurry up, one minute left!")
				cltime += showButton("Clean and back")
			}
			cltime += 60 //first waiting time
			// save time of cleaning
			save("ChastityCombination.CH", encode(padnumber(getTime())))
			setNewImage()
			show("Good. Take your chastity device and lock away this ${genital} again!")
			wait(15)
			showButton("Locked, ${keyholder}")
			def clover = Math.ceil((cltime - maxcltime*60) / 60) as int
			def ctimeadd = Math.ceil(clover / 5) as int
			if (ctimeadd > 0) {
				setNewImage()
				show("But what took you so long cleaning? You were $clover minutes late.\n\n"+
						"Therefore you get $ctimeadd hours added on your chastity duration!")
				cduration = (cduration as int) + ctimeadd * 3600
				save("ChastityCombination.CD", encode(padnumber(cduration)))
				showButton("Sorry, ${keyholder}")
			}
			show("Here's your new combination. Set each ${(ctype>1?"character":"digit")} on the lock as done before.")
			wait(5)
			block = "lockup"
			break
			
		case "exit":
			// goodbye
			setNewImage()
			show("Goodbye slave!")
			showButton("Bye ${keyholder}")
			endReached = true
			break
			
		case "":{->
			endReached = true
			}(); break

		default:
			show("Unfinished ? ("+newblock+")")
			endReached = true
		} // end switch 
	newblock = block
	} // end while 

setImage(null)
show(null)
return null

/*
 * Resources
 * scripts/ChastityCombination.groovy
 * images/ChastityCombination/info.txt
 */
